@@ -21,6 +21,7 @@
  */
 
 #include "dnsdist-proxy-protocol.hh"
+#include "dnsdist-metrics.hh"
 #include "dolog.hh"
 
 NetmaskGroup g_proxyProtocolACL;
@@ -81,13 +82,13 @@ bool handleProxyProtocol(const ComboAddress& remote, bool isTCP, const NetmaskGr
 
   ssize_t used = parseProxyHeader(query, proxyProto, realRemote, realDestination, tcp, values);
   if (used <= 0) {
-    ++g_stats.proxyProtocolInvalid;
+    ++dnsdist::metrics::g_stats.proxyProtocolInvalid;
     vinfolog("Ignoring invalid proxy protocol (%d, %d) query over %s from %s", query.size(), used, (isTCP ? "TCP" : "UDP"), remote.toStringWithPort());
     return false;
   }
   else if (static_cast<size_t>(used) > g_proxyProtocolMaximumSize) {
     vinfolog("Proxy protocol header in %s packet from %s is larger than proxy-protocol-maximum-size (%d), dropping", (isTCP ? "TCP" : "UDP"), remote.toStringWithPort(), used);
-    ++g_stats.proxyProtocolInvalid;
+    ++dnsdist::metrics::g_stats.proxyProtocolInvalid;
     return false;
   }
 
@@ -95,14 +96,14 @@ bool handleProxyProtocol(const ComboAddress& remote, bool isTCP, const NetmaskGr
 
   /* on TCP we have not read the actual query yet */
   if (!isTCP && query.size() < sizeof(struct dnsheader)) {
-    ++g_stats.nonCompliantQueries;
+    ++dnsdist::metrics::g_stats.nonCompliantQueries;
     return false;
   }
 
   if (proxyProto && g_applyACLToProxiedClients) {
     if (!acl.match(realRemote)) {
       vinfolog("Query from %s dropped because of ACL", realRemote.toStringWithPort());
-      ++g_stats.aclDrops;
+      ++dnsdist::metrics::g_stats.aclDrops;
       return false;
     }
   }
